@@ -1,12 +1,12 @@
 """Walk the directory trees and files and call the optimizers."""
+import os
 import shutil
 import time
 import traceback
 
 from multiprocessing.pool import ApplyResult, Pool
 from pathlib import Path
-from queue import SimpleQueue
-from typing import Any, Optional, Type, Union
+from typing import Optional, Type
 
 from confuse.templates import AttrDict
 from humanize import naturalsize
@@ -14,29 +14,17 @@ from termcolor import cprint
 from treestamps import Treestamps
 
 from picopt import PROGRAM_NAME
-from picopt.config import (
-    PNG_CONVERTABLE_FORMATS,
-    TIMESTAMPS_CONFIG_KEYS,
-    WEBP_CONVERTABLE_FORMATS,
-)
+from picopt.config import TIMESTAMPS_CONFIG_KEYS
 from picopt.configurable import Configurable
 from picopt.handlers.container import ContainerHandler
 from picopt.handlers.factory import create_handler
 from picopt.handlers.handler import Handler
-from picopt.handlers.image import TIFF_FORMAT, ImageHandler
+from picopt.handlers.image import ImageHandler
 from picopt.handlers.png import Png
 from picopt.handlers.webp import WebP
 from picopt.handlers.zip import CBZ, Zip
 from picopt.old_timestamps import OldTimestamps
-from picopt.stats import ReportStats
-from picopt.tasks import (
-    CompleteContainerTask,
-    CompleteDirTask,
-    CompleteTask,
-    ContainerResult,
-    DirResult,
-    Totals,
-)
+from picopt.stats import ReportStats, Totals
 
 
 class Walk(Configurable):
@@ -44,22 +32,59 @@ class Walk(Configurable):
 
     TIMESTAMPS_FILENAMES = set(Treestamps.get_filenames(PROGRAM_NAME))
 
-    def __init__(self, config: AttrDict) -> None:
-        """Initialize."""
-        super().__init__(config)
-        top_paths = []
-        for path in sorted(set(self._config.paths)):
-            if path.is_symlink() and not self._config.symlinks:
-                continue
-            top_paths.append(path)
-        self._top_paths: tuple[Path, ...] = tuple(top_paths)
-        self._timestamps: dict[Path, Treestamps] = {}
-        self._queues: dict[Path, SimpleQueue[Any]] = {}
-        if self._config.jobs:
-            self._pool = Pool(self._config.jobs)
-        else:
-            self._pool = Pool()
+    ########
+    # Init #
+    ########
+    def _convert_message(
+        self, convert_from_formats: frozenset[str], convert_handler: Type[Handler]
+    ):
+        convert_from = ", ".join(
+            sorted(convert_from_formats & frozenset(self._config.formats))
+        )
+        convert_to = convert_handler.OUTPUT_FORMAT
+        cprint(f"Converting {convert_from} to {convert_to}", "cyan")
 
+    def _init_run(self):
+        """Init Run."""
+        # Validate top_paths
+        if not self._top_paths:
+            raise ValueError("No paths to optimize.")
+        for path in self._top_paths:
+            if not path.exists():
+                raise ValueError(f"Path does not exist: {path}")
+
+        # Tell the user what we're doing
+        if self._config.verbose:
+            print("Optimizing formats:", *sorted(self._config.formats))
+            if self._config.convert_to:
+                if WebP.OUTPUT_FORMAT in self._config.convert_to:
+                    self._convert_message(self._config._convertable_formats.webp, WebP)
+                elif Png.OUTPUT_FORMAT in self._config.convert_to:
+                    self._convert_message(self._config._convertable_formats.png, Png)
+                if Zip.OUTPUT_FORMAT in self._config.convert_to:
+                    self._convert_message(frozenset([Zip.INPUT_FORMAT_RAR]), Zip)
+                if CBZ.OUTPUT_FORMAT in self._config.convert_to:
+                    self._convert_message(frozenset([CBZ.INPUT_FORMAT_RAR]), CBZ)
+            if self._config.after is not None:
+                print("Optimizing after", time.ctime(self._config.after))
+
+        # Init timestamps
+        if self._config.timestamps:
+            self._timestamps = Treestamps.map_factory(
+                self._top_paths,
+                PROGRAM_NAME,
+                self._config.verbose,
+                self._config.symlinks,
+                self._config.ignore,
+                self._config,
+                TIMESTAMPS_CONFIG_KEYS,
+            )
+            for timestamps in self._timestamps.values():
+                OldTimestamps(self._config, timestamps).import_old_timestamps()
+
+    ############
+    # Checkers #
+    ############
     def _is_skippable(self, path: Path) -> bool:
         """Handle things that are not optimizable files."""
         # File types
@@ -99,7 +124,8 @@ class Walk(Configurable):
         if self._config.after is not None:
             walk_after = self._config.after
         elif container_mtime is None:
-            walk_after = self._timestamps.get(top_path, {}).get(path)
+            timestamps = self._timestamps.get(top_path, {})
+            walk_after = timestamps.get(path)
 
         if walk_after is None:
             return False
@@ -115,16 +141,90 @@ class Walk(Configurable):
         except Exception as exc:
             cprint(str(exc), "red")
 
+    ###########
+    # Walkers #
+    ###########
+    def _finish_results(
+        self,
+        results: list[ApplyResult],
+        container_mtime: Optional[float],
+        top_path: Path,
+    ) -> None:
+        """Get the async results and total them."""
+        for result in results:
+            final_result = result.get()
+            if final_result.error:
+                cprint(final_result.error, "yellow")
+                self._totals.errors.append(final_result)
+            elif self._config.timestamps and not container_mtime:
+                timestamps = self._timestamps[top_path]
+                timestamps.set(final_result.path)
+                self._totals.bytes_in += final_result.bytes_in
+                self._totals.bytes_out += final_result.bytes_out
+
+    def walk_dir(
+        self,
+        path: Path,
+        top_path: Path,
+        container_mtime: Optional[float],
+        convert: bool,
+    ) -> None:
+        """Recursively optimize a directory."""
+        results = []
+        files = []
+        for name in sorted(os.listdir(path)):
+            entry_path = path / name
+            if entry_path.is_dir():
+                self.walk_file(entry_path, top_path, container_mtime, convert)
+            else:
+                files.append(entry_path)
+
+        for entry_path in files:
+            result = self.walk_file(entry_path, top_path, container_mtime, convert)
+            if result:
+                results.append(result)
+
+        self._finish_results(
+            results,
+            container_mtime,
+            top_path,
+        )
+
+        if self._config.timestamps and not container_mtime:
+            # Compact timestamps after every directory completes
+            timestamps = self._timestamps[top_path]
+            timestamps.set(path, compact=True)  # type: ignore
+
+    def _walk_container(self, top_path: Path, handler: ContainerHandler) -> ApplyResult:
+        """Optimize a container."""
+        result: ApplyResult
+        try:
+            handler.unpack()
+            container_mtime = handler.original_path.stat().st_mtime
+            self.walk_dir(
+                handler.tmp_container_dir,
+                top_path,
+                container_mtime,
+                handler.CONVERT,
+            )
+            result = self._pool.apply_async(handler.repack)
+        except Exception as exc:
+            traceback.print_exc()
+            args = tuple([exc])
+            result = self._pool.apply_async(handler.error, args=args)
+        return result
+
     def walk_file(
         self,
         path: Path,
         top_path: Path,
-        container_mtime: Optional[float] = None,
-        convert: bool = True,
-    ) -> Union[ApplyResult, DirResult, None]:
+        container_mtime: Optional[float],
+        convert: bool,
+    ) -> Optional[ApplyResult]:
         """Optimize an individual file."""
-        result: Union[ApplyResult, DirResult, None] = None
+        result: Optional[ApplyResult] = None
         try:
+            # START DECIDE
             if self._is_skippable(path):
                 return result
 
@@ -134,13 +234,12 @@ class Walk(Configurable):
 
             if path.is_dir():
                 if self._config.recurse or container_mtime is not None:
-                    result = self.walk_dir(
-                        path, top_path, container_mtime, convert=convert
-                    )
+                    result = self.walk_dir(path, top_path, container_mtime, convert)
                 return result
 
             if self._is_older_than_timestamp(path, top_path, container_mtime):
                 return result
+            # END DECIDE
 
             handler = create_handler(self._config, path, convert=convert)
 
@@ -153,59 +252,24 @@ class Walk(Configurable):
 
             if isinstance(handler, ContainerHandler):
                 # Unpack inline, not in the pool, and walk immediately like dirs.
-                handler.unpack()
-                result = self._walk_container_dir(top_path, handler)
+                result = self._walk_container(top_path, handler)
             elif isinstance(handler, ImageHandler):
                 result = self._pool.apply_async(handler.optimize_image)
             else:
                 raise ValueError(f"bad handler {handler}")
         except Exception as exc:
             traceback.print_exc()
-
             result = self._pool.apply_async(ReportStats, args=(path, None, exc))
         return result
 
-    def walk_dir(
-        self,
-        dir_path: Path,
-        top_path: Path,
-        container_mtime: Optional[float] = None,
-        convert: bool = True,
-    ) -> DirResult:
-        """Recursively optimize a directory."""
-        dir_result = DirResult(dir_path, [], bool(container_mtime))
-        for path in sorted(dir_path.iterdir()):
-            result = self.walk_file(path, top_path, container_mtime, convert=convert)
-            dir_result.results.append(result)
-
-        return dir_result
-
-    def _walk_container_dir(
-        self, top_path: Path, handler: ContainerHandler
-    ) -> Union[ContainerResult, ApplyResult]:
-        """Optimize a container."""
-        result: Union[ContainerResult, ApplyResult]
-        try:
-            container_mtime = handler.original_path.stat().st_mtime
-            dir_result = self.walk_dir(
-                handler.tmp_container_dir,
-                top_path,
-                container_mtime,
-                convert=handler.CONVERT,
-            )
-            result = ContainerResult(
-                handler.final_path, dir_result.results, True, handler
-            )
-        except Exception as exc:
-            args = tuple([exc])
-            result = self._pool.apply_async(handler.error, args=args)
-        return result
-
-    def _report_totals(self, totals: Totals) -> None:
+    ##########
+    # Finish #
+    ##########
+    def _report_totals(self) -> None:
         """Report the total number and percent of bytes saved."""
-        if totals.bytes_in:
-            bytes_saved = totals.bytes_in - totals.bytes_out
-            percent_bytes_saved = bytes_saved / totals.bytes_in * 100
+        if self._totals.bytes_in:
+            bytes_saved = self._totals.bytes_in - self._totals.bytes_out
+            percent_bytes_saved = bytes_saved / self._totals.bytes_in * 100
             msg = ""
             if self._config.test:
                 if percent_bytes_saved > 0:
@@ -233,132 +297,29 @@ class Walk(Configurable):
             if self._config.verbose:
                 print("Didn't optimize any files.")
 
-        if totals.errors:
+        if self._totals.errors:
             cprint("Errors with the following files:", "yellow")
-            for rs in totals.errors:
+            for rs in self._totals.errors:
                 rs.report(self._config.test, "yellow")
 
-    ######################################################################
-    #                                QUEUE                               #
-    ######################################################################
-
-    def _handle_queue_item_dir(self, top_path: Path, dir_result: DirResult):
-        """Reverse the results tree to handle directories bottom up."""
-        queue = self._queues[top_path]
-        for dir_member_result in dir_result.results:
-            if isinstance(dir_member_result, DirResult):
-                self._handle_queue_item_dir(top_path, dir_member_result)
-            else:
-                queue.put(dir_member_result)
-        if isinstance(dir_result, ContainerResult):
-            task = CompleteContainerTask(dir_result.handler)
+    ################
+    # Init and run #
+    ################
+    def __init__(self, config: AttrDict) -> None:
+        """Initialize."""
+        super().__init__(config)
+        self._totals = Totals()
+        top_paths = []
+        for path in sorted(frozenset(self._config.paths)):
+            if path.is_symlink() and not self._config.symlinks:
+                continue
+            top_paths.append(path)
+        self._top_paths: tuple[Path, ...] = tuple(top_paths)
+        self._timestamps: dict[Path, Treestamps] = {}
+        if self._config.jobs:
+            self._pool = Pool(self._config.jobs)
         else:
-            task = CompleteDirTask(dir_result.path, dir_result.in_container)
-        queue.put(task)
-
-    def _handle_queue_complete_task(self, top_path: Path, task: CompleteTask):
-        if isinstance(task, CompleteDirTask):
-            if self._config.timestamps and not task.in_container:
-                # Compact timestamps after every directory completes
-                timestamps = self._timestamps[top_path]
-                timestamps.set(task.path, compact=True)
-        elif isinstance(task, CompleteContainerTask):
-            # Repack inline, not in pool, to complete directories immediately
-            repack_result = task.handler.repack()
-            queue = self._queues[top_path]
-            queue.put(repack_result)
-        else:
-            cprint(f"WARNING: Unhandled Complete task {task}", "yellow")
-
-    def _handle_queue_item(
-        self,
-        top_path: Path,
-        totals: Totals,
-    ):
-        queue: SimpleQueue = self._queues[top_path]
-        item = queue.get()
-        if isinstance(item, ApplyResult):
-            # unpack apply results inline to preserve directory order
-            item = item.get()
-
-        if item is None:
-            pass
-        elif isinstance(item, ReportStats):
-            if item.error:
-                cprint(item.error, "yellow")
-                totals.errors.append(item)
-            else:
-                if self._config.timestamps:
-                    timestamps = self._timestamps[top_path]
-                    timestamps.set(item.path)
-                totals.bytes_in += item.bytes_in
-                totals.bytes_out += item.bytes_out
-        elif isinstance(item, DirResult):
-            self._handle_queue_item_dir(top_path, item)
-        elif isinstance(item, CompleteTask):
-            self._handle_queue_complete_task(top_path, item)
-        else:
-            cprint(f"Unhandled queue item {item}", "yellow")
-
-    ######################################################################
-    #                              END QUEUE                             #
-    ######################################################################
-
-    def _convert_message(
-        self, convert_from_formats: frozenset[str], convert_handler: Type[Handler]
-    ):
-        convert_from = ", ".join(
-            sorted(convert_from_formats & frozenset(self._config.formats))
-        )
-        convert_to = convert_handler.OUTPUT_FORMAT
-        cprint(f"Converting {convert_from} to {convert_to}", "cyan")
-
-    def _get_covertable_formats(
-        self, convertable_formats: frozenset[str]
-    ) -> frozenset[str]:
-        formats = set(convertable_formats)
-        if TIFF_FORMAT in self._config.formats:
-            formats.add(TIFF_FORMAT)
-        return frozenset(formats)
-
-    def _init_run(self):
-        """Init Run."""
-        # Validate top_paths
-        if not self._top_paths:
-            raise ValueError("No paths to optimize.")
-        for path in self._top_paths:
-            if not path.exists():
-                raise ValueError(f"Path does not exist: {path}")
-
-        # Tell the user what we're doing
-        if self._config.verbose:
-            print("Optimizing formats:", *sorted(self._config.formats))
-            if self._config.convert_to:
-                if WebP.OUTPUT_FORMAT in self._config.convert_to:
-                    formats = self._get_covertable_formats(WEBP_CONVERTABLE_FORMATS)
-                    self._convert_message(formats, WebP)
-                elif Png.OUTPUT_FORMAT in self._config.convert_to:
-                    formats = self._get_covertable_formats(PNG_CONVERTABLE_FORMATS)
-                    self._convert_message(formats, Png)
-                if Zip.OUTPUT_FORMAT in self._config.convert_to:
-                    self._convert_message(frozenset([Zip.INPUT_FORMAT_RAR]), Zip)
-                if CBZ.OUTPUT_FORMAT in self._config.convert_to:
-                    self._convert_message(frozenset([CBZ.INPUT_FORMAT_RAR]), CBZ)
-            if self._config.after is not None:
-                print("Optimizing after", time.ctime(self._config.after))
-
-        # Init timestamps
-        if self._config.timestamps:
-            self._timestamps = Treestamps.path_to_treestamps_map_factory(
-                self._top_paths,
-                PROGRAM_NAME,
-                self._config.verbose,
-                self._config.ignore,
-                self._config,
-                TIMESTAMPS_CONFIG_KEYS,
-            )
-            for timestamps in self._timestamps.values():
-                OldTimestamps(self._config, timestamps).import_old_timestamps()
+            self._pool = Pool()
 
     def run(self) -> bool:
         """Optimize all configured files."""
@@ -368,20 +329,20 @@ class Walk(Configurable):
             cprint(str(exc), "red")
             return False
 
-        # Start each queue
-        totals = Totals()
+        # Walk each top file
+        top_results = {}
         for top_path in self._top_paths:
             dirpath = Treestamps.dirpath(top_path)
-            result = self.walk_file(top_path, dirpath)
-            if dirpath not in self._queues:
-                self._queues[dirpath] = SimpleQueue()
-            queue = self._queues[dirpath]
-            self._queues[dirpath].put(result)
+            result = self.walk_file(top_path, dirpath, None, True)
+            if not result:
+                continue
+            if dirpath not in top_results:
+                top_results[dirpath] = []
+            top_results[dirpath].append(result)
 
-        # Process each queue
-        for top_path, queue in self._queues.items():
-            while not queue.empty():
-                self._handle_queue_item(top_path, totals)
+        # Finish
+        for dirpath, results in top_results.items():
+            self._finish_results(results, None, dirpath)
 
         # Shut down multiprocessing
         self._pool.close()
@@ -390,8 +351,8 @@ class Walk(Configurable):
         if self._config.timestamps:
             for top_path, timestamps in self._timestamps.items():
                 print(f"Saving timestamps for {top_path}")
-                timestamps.set(top_path, compact=True)
                 timestamps.dump()
+
         # Finish by reporting totals
-        self._report_totals(totals)
+        self._report_totals()
         return True
